from . import colors
from .application import Application
from .key import Key
from .panel import Panel
from .scrollpanel import ScrollPanel
from .inputline import InputLine
from .shellcontext import ShellContext




import locale
import curses
import sys

from . import colors
from .key import Key


class Application:
    """A generic application with curses

    Use self.screen for drawing operations.
    """
    def __init__(self):
        self.screen = None
        self.is_initialized = False
        self.preferred_encoding = None
        self._cursor_set = False
        self._original_cursor = 1

    def refresh(self, force=False):
        """Refreshes the screen

        If force is set to True, it will touch the screen first
        to enforce a full refresh"""
        if force:
            self.screen.touchwin()
        self.screen.noutrefresh()

    def size(self):
        """Refreshes the size of the screen and returns the (height,width) tuple"""
        curses.update_lines_cols()
        return self.screen.getmaxyx()

    def main(self):
        """Your entry point for the main loop"""
        raise NotImplementedError()

    def read_key(self):
        """Read the next key press from the system

        Returns a cursedspace.Key instance"""
        return Key.read(self.screen)

    def run(self):
        """The main function to call when actually running the application"""
        exception = None

        try:
            self.init_curses()
            self.main()
        except Exception as exc:
            exception = exc

        self.end_curses()

        if exception is not None:
            raise exception

        return 0

    def show_cursor(self, state):
        """Whether or not to show the cursor

        This is a wrapper for curses.curs_set, but will remember the original
        state of the cursor and restore it when the application ends, so you
        don't have to.
        """
        if isinstance(state, bool):
            state = 1 if state else 0

        try:
            previous_state = curses.curs_set(state)
        except curses.error:
            return

        if not self._cursor_set:
            self._original_cursor = previous_state

    def init_curses(self):
        """Will be called prior to using any curses operations in self.run()

        Normally you should not have to call this."""
        if self.is_initialized:
            return

        locale.setlocale(locale.LC_ALL, '')
        self.preferred_encoding = locale.getpreferredencoding()

        self.screen = curses.initscr()
        curses.def_shell_mode()

        if curses.has_colors():
            curses.start_color()
            colors.init_colors()

        curses.noecho()
        curses.cbreak()
        self.screen.keypad(True)

        curses.def_prog_mode()

        self.is_initialized = True

    def end_curses(self):
        self.screen.clear()
        self.screen.refresh()

        if self._cursor_set:
            try:
                curses.curs_set(self._original_cursor)
            except curses.error:
                pass

        try:
            curses.endwin()
        except:
            pass

    def set_term_title(self, title):
        """Attempt to set the title of the terminal emulator"""
        # trying our best
        sys.stdout.write(f'\x1b]0;{title}\x07')
        sys.stdout.write(f'\x1b]2;{title}\x07')
        sys.stdout.flush()

    def __delete__(self):
        self.end_curses()




import logging
import curses

from .panel import Panel
from .key import Key


class InputLine(Panel):
    """A panel that allows for user input as a text input field does.

    Example use case::

        # in your Application's run function:
        edit = InputLine(self, self.size()[1], (0, 0))

        while True:
            key = self.read_key()
            if key == Key.RESIZE:
                edit.resize(self.size()[1])
            elif key == Key.RETURN:
                # user accepted input, check edit.text
                break
            elif key == Key.ESCAPE:
                # user cancelled input
                break
            else:
                edit.handle_key(key)
    """
    def __init__(self, app, width, pos, text="", prefix="", background=' '):
        super().__init__(app, (1, width), pos)
        self.text = text
        self.offset = 0
        self.cursor = 0
        self.margin = 2
        self.background = background
        self.prefix = prefix
        self.read_only = False

    def paint(self, clear=False):
        self.border = Panel.BORDER_NONE
        super().paint(clear)
        self.scroll()
        visible_text = self.text[self.offset:]
        visible_text = visible_text[:self.dim[1]-1-len(self.prefix)]
        self.win.addstr(0, 0, self.background*(self.dim[1]-1))
        self.win.addstr(0, 0, self.prefix)
        self.win.addstr(0, len(self.prefix), visible_text)
        self.win.noutrefresh()

    def focus(self):
        y = 0
        x = self.cursor - self.offset + len(self.prefix)
        try:
            self.win.move(y, x)
        except curses.error:
            pass
        self.win.noutrefresh()
        return y, x

    def scroll(self):
        """Recalculate 'offset' based on window size and cursor position"""
        self.cursor = min(self.cursor, len(self.text))

        if self.margin <= self.cursor - self.offset < self.dim[1] - self.margin - len(self.prefix):
            # no need to change
            return False

        self.offset = max(0, min(len(self.text), self.cursor - self.dim[1] + self.margin + len(self.prefix)))
        return True

    def resize(self, width, compat=None):
        """Change the width of the input field"""
        if compat is not None:
            # some people might call this function with the official (height, width) signature
            # but we're ignoring the height: it's always 1!
            width = compat
        super().resize(1, width)

    def handle_key(self, key):
        must_repaint = False

        if key in [Key.RIGHT]:
            self.cursor = min(len(self.text), self.cursor+1)
            must_repaint = self.scroll()

        elif key in [Key.LEFT]:
            self.cursor = max(0, self.cursor-1)
            must_repaint = self.scroll()

        elif key in [Key.END, "^E"]:
            self.cursor = len(self.text)
            must_repaint = self.scroll()

        elif key in [Key.HOME, "^A"]:
            self.cursor = 0
            must_repaint = self.scroll()

        elif key in [Key.BACKSPACE, "^H"] and self.cursor > 0 and not self.read_only:
            self.text = self.text[:self.cursor-1] + self.text[self.cursor:]
            self.cursor -= 1
            must_repaint = True

        elif key in ["^U"] and not self.read_only:
            self.text = self.text[self.cursor:]
            self.cursor = 0
            must_repaint = True

        elif key in [Key.DELETE] and len(self.text[self.cursor:]) > 0 and not self.read_only:
            self.text = self.text[:self.cursor] + self.text[self.cursor+1:]
            must_repaint = True

        elif len(str(key)) == 1 and not self.read_only:
            self.text = self.text[:self.cursor] + str(key) + self.text[self.cursor:]
            self.cursor += 1
            must_repaint = True

        if must_repaint:
            self.paint()
            self.focus()




import curses

from .panel import Panel
from .key import Key


class ScrollPanel(Panel):
    """A generic panel that allows for scrolling content (e.g. a list)

    The content to scroll is per line in self.items.

    After updating the content of self.items you may want to reset
    self.offset and self.cursor to point to a valid element in the list of items.
    Or just set self.cursor and self.scroll to determine the self.offset
    automatically.

    ScrollPanel.SCROLL_MARGIN defines when the list should start scrolling.
    ScrollPanel.SCROLL_PAGING defines whether or not the scrolling should be paging
      instead of line by line.
    ScrollPanel.SCROLL_NEXT and ScrollPanel.SCROLL_PREVIOUS are lists of valid Keys
      that cause the selection of the next/previous item respectively. These can
      trigger scrolling.

    To control how an item is displayed, override the do_paint_item function.
    """
    SCROLL_MARGIN = 5
    SCROLL_PAGING = False
    SCROLL_NEXT = [Key.DOWN]
    SCROLL_PREVIOUS = [Key.UP]
    SCROLL_NEXT_PAGE = [Key.PGDN]
    SCROLL_PREVIOUS_PAGE = [Key.PGUP]
    SCROLL_TO_START = [Key.HOME]
    SCROLL_TO_END = [Key.END]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = []
        self.cursor = 0
        self.offset = 0
        self.list_height = 1

    @property
    def selected_item(self):
        """The currently selected item"""
        if 0 <= self.cursor < len(self.items):
            return self.items[self.cursor]
        return None

    def do_paint_item(self, y, x, maxwidth, is_selected, item):
        """Here is where you define how to paint that item on the window

        y, x: position where to draw the item
        maxwidth: available space on that line
        is_selected: whether or not that item is the currently selected one
        item: the item to paint
        """
        self.win.addstr(y, x, str(item)[:maxwidth])

    def paint(self, clear=False):
        super().paint(clear)

        for itemidx in range(self.offset, min(self.offset + self.list_height, len(self.items))):
            self.paint_item(itemidx)
        self.win.noutrefresh()

    def jump_to(self, item):
        """Select this item by instance or by index

        Puts the cursor on this item and returns True (you have to call
        paint yourself; most likely with clear=True).
        Returns False if the item is not in the list or the index was out
        of range.
        """
        if isinstance(item, int):
            if item < 0 or item >= len(self.items):
                return False
        else:
            if item not in self.items:
                return False
            item = self.items.index(item)

        self.cursor = item
        self.scroll()
        return True

    def focus(self):
        y, x = 0, 0
        if self.border & Panel.BORDER_TOP != 0:
            y += 1
        if self.border & Panel.BORDER_LEFT != 0:
            x += 1
        y += max(0, min(self.list_height-1, self.cursor - self.offset))

        try:
            self.win.move(y, x)
        except curses.error:
            pass
        self.win.noutrefresh()
        return y, x

    def resize(self, *args):
        super().resize(*args)
        self.calc_list_height()

    def calc_list_height(self):
        self.list_height = self.dim[0]
        if self.border & Panel.BORDER_TOP != 0:
            self.list_height -= 1
        if self.border & Panel.BORDER_BOTTOM != 0:
            self.list_height -= 1

    def paint_item(self, itemidx):
        if not isinstance(itemidx, int):
            itemidx = self.items.index(itemidx)
        y = itemidx - self.offset

        if y < 0 or y >= self.list_height:
            return

        x = 0
        w = self.dim[1] - 1
        if self.border & Panel.BORDER_TOP != 0:
            y += 1
        if self.border & Panel.BORDER_LEFT != 0:
            x += 1
            w -= 1
        if self.border & Panel.BORDER_RIGHT != 0:
            w -= 1

        self.do_paint_item(y, x, w, itemidx == self.cursor, self.items[itemidx])

    def scroll(self):
        self.cursor = max(0, min(self.cursor, len(self.items)-1))

        if self.SCROLL_PAGING:
            if self.cursor - self.offset >= self.list_height - self.SCROLL_MARGIN and self.list_height < len(self.items):
                self.offset = max(0, self.cursor - self.SCROLL_MARGIN)
            elif self.cursor - self.offset < self.SCROLL_MARGIN and self.offset > 0:
                self.offset = max(0, self.cursor - self.list_height + self.SCROLL_MARGIN + 1)
            else:
                return False
        else:
            if self.cursor - self.offset < self.SCROLL_MARGIN:
                self.offset = max(0, self.cursor - self.SCROLL_MARGIN)
            elif self.cursor - self.offset >= self.list_height - self.SCROLL_MARGIN:
                self.offset = min(max(0, self.cursor + self.SCROLL_MARGIN + 1 - self.list_height),
                                  max(0, len(self.items)-self.list_height))
            else:
                return False

        return True

    def handle_key(self, key):
        handled, must_repaint, must_clear = self.handle_scrolling_keys(key)

        if handled and must_repaint:
            self.paint(must_clear)

    def handle_scrolling_keys(self, key):
        """Can be called to handle scrolling keys

        If a scrolling key is found, it will likely update self.cursor.

        May call self.scroll (which updates self.offset).
        May call self.paint_item (and self.do_paint_item subsequently).

        Returns a tuple (handled, must_repaint, must_clear) meaning this:
            * handled: the key was a scrolling key
            * must_repaint: you should call self.paint
            * must_clear: you should pass clear=True to self.paint
        """
        must_repaint = False
        must_clear = False
        cursor = self.cursor
        handled = False

        if key in self.SCROLL_PREVIOUS and cursor > 0:
            cursor -= 1
            handled = True

        elif key in self.SCROLL_NEXT and cursor < len(self.items)-1:
            cursor += 1
            handled = True

        elif key in self.SCROLL_PREVIOUS_PAGE and cursor > 0:
            cursor = max(0, cursor - self.list_height)
            handled = True

        elif key in self.SCROLL_NEXT_PAGE and cursor < len(self.items)-1:
            cursor = min(len(self.items)-1, cursor + self.list_height)
            handled = True

        elif key in self.SCROLL_TO_END:
            cursor = len(self.items)-1
            handled = True

        elif key in self.SCROLL_TO_START:
            cursor = 0
            handled = True

        if self.cursor != cursor and handled:
            old_cursor = self.cursor
            self.cursor = cursor
            must_repaint = self.scroll()
            must_clear = True
            if not must_repaint:
                self.paint_item(old_cursor)
                self.paint_item(self.cursor)

        return handled, must_repaint, must_clear




import curses


class Panel:
    """A generic panel with or without border"""
    BORDER_NONE = 0
    BORDER_TOP = 1
    BORDER_BOTTOM = 2
    BORDER_LEFT = 4
    BORDER_RIGHT = 8
    BORDER_ALL = 15

    def __init__(self, app, size=None, pos=None):
        """app is the cursedspace.Application this panel belongs to
           size is a tuple (height, width)
           pos the position on the application's window (y, x)"""
        self.app = app
        self.dim = (0, 0)
        self.pos = (0, 0)
        self.win = None
        self.border = Panel.BORDER_NONE

        if size is not None:
            self.resize(*size)
            if pos is not None:
                self.move(*pos)

    def destroy(self):
        """Should be called explicitely when closing a panel to restore
        the window(s) "behind" this panel."""
        if self.win is not None:
            del self.win
            self.win = None
            # self.app.refresh(True)

    def __delete__(self):
        self.destroy()

    def handle_key(self, key):
        """Panel specific call-back when a key has been entered by the user

        key is of type cursedspace.Key"""
        pass

    def resize(self, h, w):
        """Resize the panel"""
        self.dim = (h, w)
        if self.win is None:
            self.win = curses.newwin(*self.dim, *self.pos)
        else:
            self.win.resize(*self.dim)
        self.win.clear()
        self.win.noutrefresh()

    def refresh(self, force=False):
        """Refresh the panel

        force will touch the panel's content prior to refreshing"""
        if self.win is None:
            return
        if force:
            self.win.touchwin()
        self.win.noutrefresh()

    def focus(self):
        """Move the cursor into this panel"""
        y, x = 0, 0
        if self.border & Panel.BORDER_TOP != 0:
            y += 1
        if self.border & Panel.BORDER_LEFT != 0:
            x += 1
        
        try:
            self.win.move(y, x)
        except curses.error:
            pass
        self.win.noutrefresh()

        return y, x

    def paint(self, clear=False):
        """Draws the panel into the curses buffer

        Will draw a border based on what self.border is set to.
        If clear is True, it will erase any content of the window first."""
        left = 0
        top = 0
        right = self.dim[1]-1
        bottom = self.dim[0]-1
        top_width = self.dim[1]
        bottom_width = self.dim[1]
        left_height = self.dim[0]
        right_height = self.dim[0]
        left_top = top
        right_top = top

        if clear:
            self.win.erase()

        if self.border & (Panel.BORDER_TOP + Panel.BORDER_LEFT) != 0:
            self.win.addstr(0, 0, '┌')
            left += 1
            left_height -= 1
            left_top += 1
            right_top += 1
            top_width -= 1
        if self.border & (Panel.BORDER_TOP + Panel.BORDER_RIGHT) != 0:
            self.win.addstr(0, right, '┐')
            right -= 1
            right_height -= 1
            top_width -= 1
        if self.border & (Panel.BORDER_BOTTOM + Panel.BORDER_LEFT) != 0:
            left_height -= 1
            bottom_width -= 1
            self.win.addstr(self.dim[0]-1, 0, '└')
        if self.border & (Panel.BORDER_BOTTOM + Panel.BORDER_RIGHT) != 0:
            right_height -= 1
            bottom_width -= 1
            # curses raises an exception when drawing in the lowest, most right
            # cell of the window
            try:
                self.win.addnstr(self.dim[0]-1, self.dim[1]-1, '┘', 1)
            except:
                pass
        if self.border & Panel.BORDER_TOP != 0:
            self.win.addstr(top, left, '─'*top_width)
        if self.border & Panel.BORDER_BOTTOM != 0:
            self.win.addstr(bottom, left, '─'*bottom_width)
        if self.border & Panel.BORDER_LEFT != 0:
            for y in range(left_height):
                self.win.addstr(left_top+y, 0, '│')
        if self.border & Panel.BORDER_RIGHT != 0:
            for y in range(right_height):
                self.win.addstr(right_top+y, self.dim[1]-1, '│')
        self.win.noutrefresh()

    def move(self, y, x):
        """Move the panel"""
        if self.win is None:
            return
        self.pos = (y, x)
        self.win.mvwin(y, x)
        self.app.screen.touchwin()
        self.app.screen.noutrefresh()

